#
# simulate a cellular automaton with periodic boundaries (torus-like)
# OpenMP version
#
# (c) 2024 Dorian Stoll (Julia Threads port)
# (c) 2024 Florian Werner (Julia MPI port)
# (c) 2016 Steffen Christgau (C99 port, modularization, parallelization)
# (c) 1996,1997 Peter Sanders, Ingo Boesnach (original source)
#
# command line arguments:
# #1: Number of lines
# #2: Number of iterations to be simulated
#


module ca

using FLoops
using InlineExports
using Printf

# horizontal size of the configuration
const XSIZE::Int = 1024
const LINESIZE::Int = XSIZE + 2

# annealing rule from ChoDro96 page 34
# the table is used to map the number of nonzero
# states in the neighborhood to the new state
const ANNEAL::Vector{Int8} = [0, 0, 0, 0, 1, 0, 1, 1, 1, 1]

mutable struct CellularAutomaton
    from::Matrix{Int8}
    to::Matrix{Int8}

    lines::Int
end

function init_config!(ca::CellularAutomaton)
    @ccall srand(424243::Cint)::Cvoid

    # Initialize the matrix with random values
    for y = 2:size(ca.from, 2)-1
        for x = 2:size(ca.from, 1)-1
            ca.from[x, y] = mod((@ccall rand()::Cint), 2)
        end
    end
end

function init(lines::Int)::CellularAutomaton
    ca::CellularAutomaton = CellularAutomaton(
        zeros(Int8, (LINESIZE, lines + 2)),
        zeros(Int8, (LINESIZE, lines + 2)),
        lines,
    )

    init_config!(ca)

    return ca
end

# treat torus like boundary conditions
function boundary!(ca::CellularAutomaton)
    width::Int = size(ca.from, 1)
    height::Int = size(ca.from, 2)

    ca.from[1, :] = ca.from[width-1, :]
    ca.from[width, :] = ca.from[2, :]

    ca.from[:, 1] = ca.from[:, height-1]
    ca.from[:, height] = ca.from[:, 2]
end

# computation step for each iteration
function transition(a::AbstractMatrix, x::Int, y::Int)
    #! format: off
    return ANNEAL[a[y-1, x-1] + a[y, x-1] + a[y+1, x-1] +
                  a[y-1, x  ] + a[y, x  ] + a[y+1, x  ] +
                  a[y-1, x+1] + a[y, x+1] + a[y+1, x+1] + 1]
    #! format: on
end

# make one simulation iteration.
# old configuration is in from, new one is written to to.
function simulate!(ca::CellularAutomaton)
    @floop for y = 2:size(ca.from, 2)-1
        for x = 2:size(ca.from, 1)-1
            ca.to[x, y] = transition(ca.from, y, x)
        end
    end
end

function report(buf::AbstractMatrix)
    buf[1, :] .= 0
    buf[size(buf, 1), :] .= 0

    write(stdout, buf)
end

@export function main(args::Vector{String})
    # check if Arguments are set correct
    if length(args) != 2
        println("Need 2 Arguments: number of lines and number of iterations")
        return 1
    end

    lines::Int = parse(Int, args[1])
    iterations::Int = parse(Int, args[2])

    # initializing CA
    ca::CellularAutomaton = init(lines)

    for _ = 1:iterations
        boundary!(ca)
        simulate!(ca)

        ca.from, ca.to = ca.to, ca.from
    end

    report(ca.from[:, 2:ca.lines+1])
end

end
