module boundary
   use, intrinsic :: iso_fortran_env
   use, intrinsic :: iso_c_binding

   use grid

   implicit none

   ! Enum defining the two implemented boundary conditions.
   enum, bind(C)
      enumerator :: BC_TYPE_CLOSED
      enumerator :: BC_TYPE_CONSTANT
   end enum

   ! Enum defining all 4 possible sides to a 1D and 2D grid.
   enum, bind(C)
      enumerator :: BC_SIDE_LEFT = 1
      enumerator :: BC_SIDE_RIGHT = 2
      enumerator :: BC_SIDE_TOP = 3
      enumerator :: BC_SIDE_BOTTOM = 4
   end enum

   integer, parameter :: BC_TYPE = kind(BC_TYPE_CONSTANT)
   integer, parameter :: BC_SIDE = kind(BC_SIDE_BOTTOM)

   ! This class defines the boundary conditions of individual boundary elements.
   ! These can be flexibly used and combined later in other classes.
   ! The class serves as an auxiliary class for structuring the Boundary class.
   type :: boundaryelement_t
      integer(BC_TYPE) :: type
      real(real64) :: value
   end type

   type :: boundaryside_t
      ! Vector with Boundary Element information
      type(boundaryelement_t), dimension(:), allocatable :: data
   end type

   type :: boundary_t
      integer(int8) :: dims
      integer(int32) :: cols
      integer(int32) :: rows

      type(boundaryside_t), dimension(4) :: boundaries
   end type

contains
   ! Construct a new Boundary Element object for the closed case. The boundary type is here
   ! automatically set to the type BC_TYPE_CLOSED, where the value takes -1 and does not hold
   ! any physical meaning.
   subroutine initBoundaryElementClosed(this)
      type(boundaryelement_t), intent(out) :: this

      this%type = BC_TYPE_CLOSED
      this%value = -1
   end subroutine

   ! Construct a new Boundary Element object for the constant case. The boundary type is
   ! automatically set to the type BC_TYPE_CONSTANT.
   !
   ! Arguments:
   !   value: Value of the constant concentration to be assumed at the corresponding
   !          boundary element.
   subroutine initBoundaryElementConstant(this, value)
      type(boundaryelement_t), intent(out) :: this
      real(real64), intent(in) :: value

      this%type = BC_TYPE_CONSTANT
      this%value = value
   end subroutine

   ! Allows changing the boundary type of a corresponding BoundaryElement object.
   !
   ! Arguments:
   !   type: Type of boundary condition. Either BC_TYPE_CONSTANT or BC_TYPE_CLOSED.
   subroutine setBoundaryElementType(this, type)
      type(boundaryelement_t), intent(inout) :: this
      integer(BC_TYPE), intent(in) :: type

      this%type = type
   end subroutine

   ! Sets the value of a boundary condition for the constant case.
   !
   ! Arguments:
   !   value: Concentration to be considered constant for the corresponding boundary element.
   subroutine setBoundaryElementValue(this, value)
      type(boundaryelement_t), intent(inout) :: this
      real(real64), intent(in) :: value

      this%value = value
   end subroutine

   ! Return the type of the boundary condition, i.e. whether the boundary is considered closed or constant.
   !
   ! Returns: Type of boundary condition, either BC_TYPE_CLOSED or BC_TYPE_CONSTANT.
   subroutine getBoundaryElementType(this, result)
      type(boundaryelement_t), intent(in) :: this
      integer(BC_TYPE), intent(out) :: result

      result = this%type
   end subroutine

   ! Return the concentration value for the constant boundary condition.
   !
   ! Returns: Value of the concentration.
   subroutine getBoundaryElementValue(this, result)
      type(boundaryelement_t), intent(in) :: this
      real(real64), intent(out) :: result

      result = this%value
   end subroutine

   subroutine initBoundarySide(this, size)
      type(boundaryside_t), intent(out) :: this
      integer, intent(in) :: size

      type(boundaryelement_t) :: source
      call initBoundaryElementClosed(source)

      allocate(this%data(size), source=source)
   end subroutine

   ! Creates a boundary object for a 1D grid
   !
   ! Arguments:
   !   length: Length of the grid
   subroutine initBoundary1D(this, length)
      type(boundary_t), intent(out) :: this
      integer(int32), intent(in) :: length

      type(grid_t) :: grid
      call initGrid1D(grid, length)

      call initBoundaryFromGrid(this, grid)
   end

   ! Creates a boundary object for a 2D grid
   !
   ! Arguments:
   !   rows: Number of rows of the grid
   !   cols: Number of columns of the grid
   subroutine initBoundary2D(this, rows, cols)
      type(boundary_t), intent(out) :: this
      integer(int32), intent(in) :: rows
      integer(int32), intent(in) :: cols

      type(grid_t) :: grid
      call initGrid2D(grid, rows, cols)

      call initBoundaryFromGrid(this, grid)
   end subroutine

   ! Creates a boundary object based on the passed grid object and initializes the boundaries as closed.
   !
   ! Arguments:
   !   grid: Grid object on the basis of which the simulation takes place and from which the
   !         dimensions (in 2D case) are taken.
   subroutine initBoundaryFromGrid(this, grid)
      type(boundary_t), intent(out) :: this
      type(grid_t), intent(in) :: grid

      integer :: temp

      type(boundaryelement_t) :: source
      call initBoundaryElementClosed(source)

      call getGridDim(grid, temp)
      this%dims = INT(temp, kind=int8)

      call getGridCol(grid, temp)
      this%cols = INT(temp, kind=int32)

      call getGridROw(grid, temp)
      this%rows = INT(temp, kind=int32)

      if (this%dims == 1) then
         allocate(this%boundaries(BC_SIDE_LEFT)%data(1), source=source)
         allocate(this%boundaries(BC_SIDE_RIGHT)%data(1), source=source)
      else if (this%dims == 2) then
         allocate(this%boundaries(BC_SIDE_LEFT)%data(this%rows), source=source)
         allocate(this%boundaries(BC_SIDE_RIGHT)%data(this%rows), source=source)
         allocate(this%boundaries(BC_SIDE_TOP)%data(this%cols), source=source)
         allocate(this%boundaries(BC_SIDE_BOTTOM)%data(this%cols), source=source)
      end if
   end subroutine

   ! Sets all elements of the specified boundary side to the boundary condition closed.
   !
   ! Arguments:
   !   side: Side to be set to closed, e.g. BC_SIDE_LEFT.
   subroutine setBoundarySideClosed(this, side)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side

      type(boundaryelement_t) :: source

      if (this%dims == 1) then
         if (side == BC_SIDE_BOTTOM .or. side == BC_SIDE_TOP) then
            error stop "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist."
         end if
      end if

      call initBoundaryElementClosed(source)
      this%boundaries(side)%data(:) = source
   end subroutine

   ! Sets all elements of the specified boundary side to the boundary condition constant.
   ! Thereby the concentration values of the boundaries are set to the passed value.
   !
   ! Arguments:
   !   side: Side to be set to constant, e.g. BC_SIDE_LEFT.
   !   value: Concentration to be set for all elements of the specified page.
   subroutine setBoundarySideConstant(this, side, value)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side
      real(real64), intent(in) :: value

      type(boundaryelement_t) :: source

      if (this%dims == 1) then
         if (side == BC_SIDE_BOTTOM .or. side == BC_SIDE_TOP) then
            error stop "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist."
         end if
      end if

      call initBoundaryElementConstant(source, value)
      this%boundaries(side)%data(:) = source
   end subroutine

   ! Specifically sets the boundary element of the specified side defined by the index to the
   ! boundary condition closed.
   !
   ! Arguments:
   !   side: Side in which an element is to be defined as closed.
   !   index: Index of the boundary element on the corresponding boundary side.
   !          Must index an element of the corresponding side.
   subroutine setBoundaryElementClosed(this, side, index)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side
      integer, intent(in) :: index

      ! tests whether the index really points to an element of the boundary side.
      if (size(this%boundaries(side)%data, 1) < index .or. index < 1) then
         error stop "Index is selected either too large or too small."
      end if

      call setBoundaryElementType(this%boundaries(side)%data(index), BC_TYPE_CLOSED)
   end subroutine

   ! Specifically sets the boundary element of the specified side defined by the index to the boundary
   ! condition constant with the given concentration value.
   !
   ! Arguments:
   !   side: Side in which an element is to be defined as constant.
   !   index: Index of the boundary element on the corresponding boundary side.
   !          Must index an element of the corresponding side.
   !   value: Concentration value to which the boundary element should be set.
   subroutine setBoundaryElementConstant(this, side, index, value)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side
      integer, intent(in) :: index
      real(real64), intent(in) :: value

      ! tests whether the index really points to an element of the boundary side.
      if (size(this%boundaries(side)%data, 1) < index .or. index < 1) then
         error stop "Index is selected either too large or too small."
      end if

      call setBoundaryElementType(this%boundaries(side)%data(index), BC_TYPE_CONSTANT)
      call setBoundaryElementValue(this%boundaries(side)%data(index), value)
   end subroutine

   ! Returns the boundary condition of a specified side as a vector of BoundarsElement objects.
   !
   ! Arguments:
   !   side: Boundary side from which the boundary conditions are to be returned.
   ! Returns: Contains the boundary conditions as BoundaryElement<T> objects.
   subroutine getBoundarySide(this, side, result)
      type(boundary_t), intent(inout), target :: this
      integer(BC_SIDE), intent(in) :: side
      type(boundaryside_t), intent(out), pointer :: result

      if (this%dims == 1) then
         if (side == BC_SIDE_BOTTOM .or. side == BC_SIDE_TOP) then
            error stop "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist."
         end if
      end if

      result => this%boundaries(side)
   end subroutine

   ! Returns the boundary condition of a specified element on a given side.
   !
   ! Arguments:
   !   side: Boundary side in which the boundary condition is located.
   !   index: Index of the boundary element on the corresponding boundary side.
   !          Must index an element of the corresponding side.
   ! Return: Boundary condition as a BoundaryElement<T> object.
   subroutine getBoundaryElement(this, side, index, result)
      type(boundary_t), intent(inout), target :: this
      integer(BC_SIDE), intent(in) :: side
      integer, intent(in) :: index
      type(boundaryelement_t), intent(out), pointer :: result

      ! tests whether the index really points to an element of the boundary side.
      if (size(this%boundaries(side)%data, 1) < index .or. index < 1) then
         error stop "Index is selected either too large or too small."
      end if

      result => this%boundaries(side)%data(index)
   end subroutine

   ! Returns the type of a boundary condition, i.e. either BC_TYPE_CLOSED or BC_TYPE_CONSTANT.
   !
   ! Arguments:
   !   side: Boundary side in which the boundary condition is located.
   !   index: Index of the boundary element on the corresponding boundary side.
   !          Must index an element of the corresponding side.
   ! Return: Boundary Type of the corresponding boundary condition.
   subroutine getBoundaryElementType2(this, side, index, result)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side
      integer, intent(in) :: index
      integer(BC_TYPE), intent(out) :: result

      ! tests whether the index really points to an element of the boundary side.
      if (size(this%boundaries(side)%data, 1) < index .or. index < 1) then
         error stop "Index is selected either too large or too small."
      end if

      call getBoundaryElementType(this%boundaries(side)%data(index), result)
   end subroutine

   ! Returns the concentration value of a corresponding BoundaryElement<T>
   ! object if it is a constant boundary condition.
   !
   ! Arguments:
   !   side: Boundary side in which the boundary condition is located.
   !   index: Index of the boundary element on the corresponding boundary side.
   !          Must index an element of the corresponding side.
   ! Return: Concentration of the corresponding BoundaryElement<T> object.
   subroutine getBoundaryElementValue2(this, side, index, result)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side
      integer, intent(in) :: index
      real(real64), intent(out) :: result

      ! tests whether the index really points to an element of the boundary side.
      if (size(this%boundaries(side)%data, 1) < index .or. index < 1) then
         error stop "Index is selected either too large or too small."
      end if

      call getBoundaryElementValue(this%boundaries(side)%data(index), result)
   end subroutine

   ! Get thes Boundary Side Values as a vector. Value is -1 in case some specific boundary is closed.
   !
   ! Arguments:
   !   side: Boundary side for which the values are to be returned.
   ! Returns: Vector with values as T.
   subroutine getBoundarySideValues(this, side, result)
      type(boundary_t), intent(inout) :: this
      integer(BC_SIDE), intent(in) :: side
      real(real64), dimension(:), intent(out), allocatable :: result

      integer :: length
      integer :: i

      integer(BC_TYPE) :: type

      length = size(this%boundaries(side)%data, 1)
      allocate(result(length))

      do i = 1, length
         call getBoundaryElementType2(this, side, i, type)

         if (type == BC_TYPE_CLOSED) then
            result(i) = -1
            continue
         end if

         call getBoundaryElementValue2(this, side, i, result(i))
      end do
   end subroutine
end module
