module btcs
   use, intrinsic :: iso_c_binding
   use, intrinsic :: iso_fortran_env

   use boundary
   use grid
   use utils

   implicit none

   type :: diagonals_t
      real(real64), dimension(:), allocatable :: left
      real(real64), dimension(:), allocatable :: center
      real(real64), dimension(:), allocatable :: right
   end type

contains
   ! calculates coefficient for boundary in constant case
   subroutine calcBoundaryCoeffConstant(alpha_center, alpha_side, sx, result)
      real(real64), intent(in) :: alpha_center
      real(real64), intent(in) :: alpha_side
      real(real64), intent(in) :: sx
      real(real64), dimension(2), intent(out) :: result

      result(1) = 1 + sx * (calcAlphaIntercell(alpha_center, alpha_side) + 2 * alpha_center)
      result(2) = -sx * calcAlphaIntercell(alpha_center, alpha_side)
   end subroutine

   ! calculates coefficient for boundary in closed case
   subroutine calcBoundaryCoeffClosed(alpha_center, alpha_side, sx, result)
      real(real64), intent(in) :: alpha_center
      real(real64), intent(in) :: alpha_side
      real(real64), intent(in) :: sx
      real(real64), dimension(2), intent(out) :: result

      result(1) = 1 + sx * calcAlphaIntercell(alpha_center, alpha_side)
      result(2) = -sx * calcAlphaIntercell(alpha_center, alpha_side)
   end subroutine

   ! creates coefficient matrix for next time step from alphas in x-direction
   subroutine createCoeffMatrix(alpha, bcLeft, bcRight, numCols, rowIndex, sx, result)
      real(real64), dimension(:, :), intent(in) :: alpha
      type(boundaryside_t), intent(in) :: bcLeft
      type(boundaryside_t), intent(in) :: bcRight
      integer, intent(in) :: numCols
      integer, intent(in) :: rowIndex
      real(real64), intent(in) :: sx
      type(diagonals_t), intent(out) :: result

      integer(BC_TYPE) :: type = BC_TYPE_CLOSED
      real(real64), dimension(2) :: boundaryCoeff

      integer :: n
      integer :: i

      allocate(result%left(numCols), source=0.0_real64)
      allocate(result%center(numCols), source=0.0_real64)
      allocate(result%right(numCols), source=0.0_real64)

      ! left column
      call getBoundaryElementType(bcLeft%data(rowIndex), type)

      if (type .eq. BC_TYPE_CONSTANT) then
         call calcBoundaryCoeffConstant(alpha(1, rowIndex), alpha(2, rowIndex), sx, boundaryCoeff)

         result%center(1) = boundaryCoeff(1)
         result%right(1) = boundaryCoeff(2)
      else if (type .eq. BC_TYPE_CLOSED) then
         call calcBoundaryCoeffClosed(alpha(1, rowIndex), alpha(2, rowIndex), sx, boundaryCoeff)

         result%center(1) = boundaryCoeff(1)
         result%right(1) = boundaryCoeff(2)
      else
         error stop "Undefined Boundary Condition Type somewhere on Left or Top!"
      end if

      ! inner columns
      n = numCols - 1

      do i = 2, n
         result%left(i) = -sx * calcAlphaIntercell(alpha(i-1, rowIndex), alpha(i, rowIndex))
         result%center(i) = &
            1 + &
            sx * ( &
            calcAlphaIntercell(alpha(i, rowIndex), alpha(i+1, rowIndex)) + &
            calcAlphaIntercell(alpha(i-1, rowIndex), alpha(i, rowIndex)) &
            )
            result%right(i) = -sx * calcAlphaIntercell(alpha(i, rowIndex), alpha(i+1, rowIndex))
      end do

      ! right column
      call getBoundaryElementType(bcRight%data(rowIndex), type)

      if (type .eq. BC_TYPE_CONSTANT) then
         call calcBoundaryCoeffConstant(alpha(numCols, rowIndex), alpha(numCols-1, rowIndex), sx, boundaryCoeff)

         result%center(numCols) = boundaryCoeff(1)
         result%left(numCols) = boundaryCoeff(2)
      else if (type .eq. BC_TYPE_CLOSED) then
         call calcBoundaryCoeffClosed(alpha(numCols, rowIndex), alpha(numCols-1, rowIndex), sx, boundaryCoeff)

         result%center(numCols) = boundaryCoeff(1)
         result%left(numCols) = boundaryCoeff(2)
      else
         error stop "Undefined Boundary Condition Type somewhere on Right or Bottom!"
      end if
   end subroutine

   ! calculates explicit concentration at boundary in closed case
   subroutine calcExplicitConcentrationsBoundaryClosed(conc_center, alpha_center, alpha_neigbor, sy, result)
      real(real64), intent(in) :: conc_center
      real(real64), intent(in) :: alpha_center
      real(real64), intent(in) :: alpha_neigbor
      real(real64), intent(in) :: sy
      real(real64), intent(out) :: result

      result = sy * calcAlphaIntercell(alpha_center, alpha_neigbor) * conc_center + (1 - sy * (calcAlphaIntercell(alpha_center, alpha_neigbor))) * conc_center
   end subroutine

   ! calculates explicity concentration at boundary in constant case
   subroutine calcExplicitConcentrationsBoundaryConstant(conc_center, conc_bc, alpha_center, alpha_neighbor, sy, result)
      real(real64), intent(in) :: conc_center
      real(real64), intent(in) :: conc_bc
      real(real64), intent(in) :: alpha_center
      real(real64), intent(in) :: alpha_neighbor
      real(real64), intent(in) :: sy
      real(real64), intent(out) :: result

      real(real64) :: inter_cell

      inter_cell = calcAlphaIntercell(alpha_center, alpha_neighbor)
      result = sy * inter_cell * conc_center + (1 - sy * (inter_cell + alpha_center)) * conc_center + sy * alpha_center * conc_bc
   end subroutine

   subroutine createSolutionVector(concentrations, alphaX, alphaY, bcLeft, bcRight, bcTop, bcBottom, rowIndex, sx, sy, result)
      real(real64), dimension(:, :), intent(in) :: concentrations
      real(real64), dimension(:, :), intent(in) :: alphaX
      real(real64), dimension(:, :), intent(in) :: alphaY

      type(boundaryside_t), intent(in) :: bcLeft
      type(boundaryside_t), intent(in) :: bcRight
      type(boundaryside_t), intent(in) :: bcTop
      type(boundaryside_t), intent(in) :: bcBottom

      integer, intent(in) :: rowIndex

      real(real64), intent(in) :: sx
      real(real64), intent(in) :: sy

      real(real64), dimension(:), intent(out) :: result

      integer :: numRows
      integer :: i

      integer(BC_TYPE) :: type
      real(real64) :: value

      numRows = size(concentrations, 2)

      type = BC_TYPE_CLOSED

      ! inner rows
      if (rowIndex .gt. 1 .and. rowIndex .lt. numRows) then
         do i = 1, size(result)
            result(i) =  &
               sy *  &
               calcAlphaIntercell(alphaY(i, rowIndex), alphaY(i, rowIndex+1)) * &
               concentrations(i, rowIndex+1) + &
               ( &
               1 - &
               sy * ( &
               calcAlphaIntercell(alphaY(i, rowIndex), alphaY(i, rowIndex+1)) + &
               calcAlphaIntercell(alphaY(i, rowIndex-1), alphaY(i, rowIndex)) &
               ) &
               ) * concentrations(i, rowIndex) + &
               sy * &
               calcAlphaIntercell(alphaY(i, rowIndex-1), alphaY(i, rowIndex)) * &
               concentrations(i, rowIndex-1)
         end do
      else if (rowIndex .eq. 1) then ! first row
         do i = 1, size(result)
            call getBoundaryElementType(bcTop%data(i), type)

            if (type .eq. BC_TYPE_CONSTANT) then
               call getBoundaryElementValue(bcTop%data(i), value)
               call calcExplicitConcentrationsBoundaryConstant( &
                  concentrations(i, rowIndex), &
                  value, &
                  alphaY(i, rowIndex), &
                  alphaY(i, rowIndex+1), &
                  sy, &
                  result(i))
            else if (type .eq. BC_TYPE_CLOSED) then
               call calcExplicitConcentrationsBoundaryClosed( &
                  concentrations(i, rowIndex), &
                  alphaY(i, rowIndex), &
                  alphaY(i, rowIndex+1), &
                  sy, &
                  result(i))
            else
               error stop "Undefined Boundary Condition Type somewhere on Left or Top!"
            end if
         end do
      else if (rowIndex .eq. numRows) then ! last row
         do i = 1, size(result)
            call getBoundaryElementType(bcBottom%data(i), type)

            if (type .eq. BC_TYPE_CONSTANT) then
               call getBoundaryElementValue(bcBottom%data(i), value)
               call calcExplicitConcentrationsBoundaryConstant( &
                  concentrations(i, rowIndex), &
                  value, &
                  alphaY(i, rowIndex), &
                  alphaY(i, rowIndex-1), &
                  sy, &
                  result(i))
            else if (type .eq. BC_TYPE_CLOSED) then
               call calcExplicitConcentrationsBoundaryClosed( &
                  concentrations(i, rowIndex), &
                  alphaY(i, rowIndex), &
                  alphaY(i, rowIndex-1), &
                  sy, &
                  result(i))
            else
               error stop "Undefined Boundary Condition Type somewhere on Right or Bottom!"
            end if
         end do
      end if

      call getBoundaryElementType(bcLeft%data(rowIndex), type)

      ! first column -> additional fixed concentration change from perpendicular dimension in constant bc case
      if (type .eq. BC_TYPE_CONSTANT) then
         call getBoundaryElementValue(bcLeft%data(rowIndex), value)
         result(1) = result(1) + (2 * sx * alphaX(1, rowIndex) * value)
      end if

      call getBoundaryElementType(bcRight%data(rowIndex), type)

      ! last column -> additional fixed concentration change from perpendicular dimension in constant bc case
      if (type .eq. BC_TYPE_CONSTANT) then
         call getBoundaryElementValue(bcRight%data(rowIndex), value)
         result(size(result)) = result(size(result)) + (2 * sx * alphaX(size(result), rowIndex) * value)
      end if
   end subroutine

   ! solver for linear equation system; A corresponds to coefficient matrix, b to the solution vector
   ! implementation of Thomas Algorithm
   subroutine ThomasAlgorithm(A, x_vec)
      type(diagonals_t), intent(inout) :: A
      real(real64), dimension(:), intent(inout) :: x_vec

      integer :: n
      integer :: i

      n = size(x_vec, 1)

      ! start solving - c_diag and x_vec are overwritten
      n = n - 1

      A%right(1) = A%right(1) / A%center(1)
      x_vec(1) = x_vec(1) / A%center(1)

      do i = 2, n
         A%right(i) = A%right(i) / (A%center(i) - A%left(i) * A%right(i-1))
         x_vec(i) = (x_vec(i) - A%left(i) * x_vec(i-1)) / (A%center(i) - A%left(i) * A%right(i-1))
      end do

      x_vec(n+1) = (x_vec(n+1) - A%left(n+1) * x_vec(n)) / (A%center(n+1) - A%left(n+1) * A%right(n))

      do i = 0, n-1
         x_vec(n-i) = x_vec(n-i) - A%right(n-i) * x_vec(n-i+1)
      end do
   end subroutine

   ! BTCS solution for 2D grid
   subroutine BTCS_2D(grid, bc, timestep)
      type(grid_t), intent(inout) :: grid
      type(boundary_t), intent(inout) :: bc
      real(real64), intent(in) :: timestep

      integer :: rowMax
      integer :: colMax

      real(real64) :: deltaCol
      real(real64) :: deltaRow

      real(real64) :: sx
      real(real64) :: sy

      real(real64), dimension(:, :), pointer :: concentrations
      real(real64), dimension(:, :), allocatable :: concentrations_t1

      real(real64), dimension(:, :), pointer :: alphaX
      real(real64), dimension(:, :), pointer :: alphaY

      type(boundaryside_t), pointer :: bcLeft
      type(boundaryside_t), pointer :: bcRight
      type(boundaryside_t), pointer :: bcTop
      type(boundaryside_t), pointer :: bcBottom

      integer :: i

      type(diagonals_t) :: A
      real(real64), dimension(:), allocatable :: b

      call getGridRow(grid, rowMax)
      call getGridCol(grid, colMax)

      call getGridDeltaCol(grid, deltaCol)
      call getGridDeltaRow(grid, deltaRow)

      sx = timestep / (2 * deltaCol * deltaCol)
      sy = timestep / (2 * deltaRow * deltaRow)

      allocate(concentrations_t1(colMax, rowMax))

      call getGridAlphaX(grid, alphaX)
      call getGridAlphaY(grid, alphaY)

      call getBoundarySide(bc, BC_SIDE_LEFT, bcLeft)
      call getBoundarySide(bc, BC_SIDE_RIGHT, bcRight)
      call getBoundarySide(bc, BC_SIDE_TOP, bcTop)
      call getBoundarySide(bc, BC_SIDE_BOTTOM, bcBottom)

      call getGridConcentrations(grid, concentrations)

      allocate(b(colMax))

      !$omp parallel do private(A, b)
      do i = 1, rowMax
         call createCoeffMatrix(alphaX, bcLeft, bcRight, colMax, i, sx, A)
         call createSolutionVector(concentrations, alphaX, alphaY, bcLeft, bcRight, bcTop, bcBottom, i, sx, sy, b)

         call ThomasAlgorithm(A, b)
         concentrations_t1(:, i) = b
      end do

      concentrations_t1(:, :) = transpose(concentrations_t1)
      alphaX(:, :) = transpose(alphaX)
      alphaY(:, :) = transpose(alphaY)

      deallocate(b)
      allocate(b(rowMax))

      !$omp parallel do private(A, b)
      do i = 1, colMax
         ! swap alphas, boundary conditions and sx/sy for column-wise calculation
         call createCoeffMatrix(alphaY, bcTop, bcBottom, rowMax, i, sy, A)
         call createSolutionVector(concentrations_t1, alphaY, alphaX, bcTop, bcBottom, bcLeft, bcRight, i, sy, sx, b)

         call ThomasAlgorithm(A, b)
         concentrations(i, :) = b
      end do
   end subroutine

   ! differentiate between 1D and 2D grid
   subroutine BTCS_Thomas(grid, bc, timestep)
      type(grid_t), intent(inout) :: grid
      type(boundary_t), intent(inout) :: bc
      real(real64), intent(in) :: timestep

      integer :: dim
      call getGridDim(grid, dim)

      if (dim .eq. 2) then
         call BTCS_2D(grid, bc, timestep)
      else
         error stop "Error: Only 2-dimensional grids are defined!"
      end if
   end subroutine
end module
