module grid
   use, intrinsic :: iso_fortran_env
   use, intrinsic :: iso_c_binding

   implicit none

   real(real64), parameter :: MAT_INIT_VALUE = 0

   ! Holds a matrix with concenctration and respective matrix/matrices of alpha coefficients.
   type :: grid_t
      integer :: col
      integer :: row
      integer :: dim

      real(real64) :: domainCol
      real(real64) :: domainRow

      real(real64) :: deltaCol
      real(real64) :: deltaRow

      real(real64), dimension(:, :), allocatable :: concentrations
      real(real64), dimension(:, :), allocatable :: alphaX
      real(real64), dimension(:, :), allocatable :: alphaY
   end type

contains
   ! Constructs a new 1D-Grid object of a given length, which holds a matrix with concentrations
   ! and a respective matrix of alpha coefficients. The domain length is per default the same as
   ! the length. The concentrations are all 20 by default and the alpha coefficients are 1.
   !
   ! Arguments:
   !   length: Length of the 1D-Grid. Must be greater than 3.
   subroutine initGrid1D(this, length)
      integer, intent(in) :: length
      type(grid_t), intent(out) :: this

      if (length .le. 3) then
         error stop "Given grid length too small. Must be greater than 3."
      end if

      this%col = length
      this%row = 1

      this%dim = 1

      this%domainCol = real(length, kind=real64)
      this%domainRow = 0

      this%deltaCol = this%domainCol / real(this%col, kind=real64) ! -> 1
      this%deltaRow = 0

      allocate(this%concentrations(this%col, 1), source=MAT_INIT_VALUE)
      allocate(this%alphaX(this%col, 1), source=MAT_INIT_VALUE)
   end subroutine

   ! Constructs a new 2D-Grid object of given dimensions, which holds a matrix with concentrations
   ! and the respective matrices of alpha coefficient for each direction. The domain in
   ! x- and y-direction is per default equal to the col length and row length, respectively. The
   ! concentrations are all 20 by default across the entire grid and the alpha coefficients 1 in
   ! both directions.
   !
   ! Arguments:
   !   row: Length of the 2D-Grid in y-direction. Must be greater than 3.
   !   col: Length of the 2D-Grid in x-direction. Must be greater than 3.
   subroutine initGrid2D(this, row, col)
      integer, intent(in) :: row
      integer, intent(in) :: col
      type(grid_t), intent(out) :: this

      if ((row .le. 1) .or. (col .le. 1)) then
         error stop "At least one dimension is 1. Use 1D grid for better results."
      end if

      this%col = col
      this%row = row

      this%dim = 2

      this%domainCol = real(this%col, kind=real64)
      this%domainRow = real(this%row, kind=real64)

      this%deltaCol = this%domainCol / real(this%col, kind=real64)
      this%deltaRow = this%domainRow / real(this%row, kind=real64)

      allocate(this%concentrations(col, row), source=MAT_INIT_VALUE)
      allocate(this%alphaX(col, row), source=MAT_INIT_VALUE)
      allocate(this%alphaY(col, row), source=MAT_INIT_VALUE)
   end subroutine

   ! Sets the concentrations matrix for a 1D or 2D-Grid.
   !
   ! Arguments:
   !   concentrations: An Eigen3 MatrixX<T> holding the concentrations.
   !                   Matrix must have correct dimensions as defined in row and col.
   !                   (Or length, in 1D case).
   subroutine setGridConcentrations(this, concentrations)
      type(grid_t), intent(inout) :: this
      real(real64), dimension(:, :), intent(in) :: concentrations

      this%concentrations(:, :) = concentrations(:, :)
   end subroutine

   ! Gets the concentrations matrix for a Grid.
   !
   ! Returns: A matrix holding the concentrations and having the same dimensions as the grid.
   subroutine getGridConcentrations(this, result)
      type(grid_t), intent(in), target :: this
      real(real64), dimension(:, :), intent(out), pointer :: result

      result => this%concentrations
   end subroutine

   ! Set the alpha coefficients of a 1D-Grid. Grid must be one dimensional.
   !
   ! Arguments
   !   alpha: A matrix with 1 row holding the alpha coefficients. Matrix columns must have
   !          same size as length of grid.
   subroutine setGridAlpha1D(this, alpha)
      type(grid_t), intent(inout) :: this
      real(real64), dimension(:, :), intent(in) :: alpha

      if (this%dim .ne. 1) then
         error stop "Grid is not one dimensional, you should probably use 2D setter function!"
      end if

      if (size(this%alphaX) .ne. size(alpha)) then
         error stop "Given matrix of alpha coefficients mismatch with Grid dimensions!"
      end if

      this%alphaX(:, :) = alpha(:, :)
   end subroutine

   ! Set the alpha coefficients of a 2D-Grid. Grid must be two dimensional.
   !
   ! Arguments
   !   alphaX: A matrix holding the alpha coefficients in x-direction. Matrix must be of
   !           same size as the grid.
   !   alphaY: A matrix holding the alpha coefficients in y-direction. Matrix must be of
   !           same size as the grid.
   subroutine setGridAlpha2D(this, alphaX, alphaY)
      type(grid_t), intent(inout) :: this
      real(real64), dimension(:, :), intent(in) :: alphaX
      real(real64), dimension(:, :), intent(in) :: alphaY

      if (this%dim .ne. 2) then
         error stop "Grid is not two dimensional, you should probably use 1D setter function!"
      end if

      if (size(alphaX) .ne. size(this%alphaX)) then
         error stop "Given matrix of alpha coefficients in x-direction mismatch with Grid dimensions!"
      end if

      if (size(alphaY) .ne. size(this%alphaY)) then
         error stop "Given matrix of alpha coefficients in y-direction mismatch with Grid dimensions!"
      end if

      this%alphaX(:, :) = alphaX(:, :)
      this%alphaY(:, :) = alphaY(:, :)
   end subroutine

   ! Gets the matrix of alpha coefficients of a 1D-Grid. Grid must be one dimensional.
   !
   ! Returns: A matrix with 1 row holding the alpha coefficients.
   subroutine getGridAlpha(this, result)
      type(grid_t), intent(in), target :: this
      real(real64), dimension(:, :), intent(out), pointer :: result

      if (this%dim .ne. 1) then
         error stop "Grid is not one dimensional, you should probably use either getAlphaX() or getAlphaY()!"
      end if

      result => this%alphaX
   end subroutine

   ! Gets the matrix of alpha coefficients in x-direction of a 2D-Grid. Grid must be two dimensional.
   !
   ! Returns: A matrix holding the alpha coefficients in x-direction.
   subroutine getGridAlphaX(this, result)
      type(grid_t), intent(in), target :: this
      real(real64), dimension(:, :), intent(out), pointer :: result

      if (this%dim .ne. 2) then
         error stop "Grid is not two dimensional, you should probably use getAlpha()!"
      end if

      result => this%alphaX
   end subroutine

   ! Gets the matrix of alpha coefficients in y-direction of a 2D-Grid. Grid must be two dimensional.
   !
   ! Returns: A matrix holding the alpha coefficients in y-direction.
   subroutine getGridAlphaY(this, result)
      type(grid_t), intent(in), target :: this
      real(real64), dimension(:, :), intent(out), pointer :: result

      if (this%dim .ne. 2) then
         error stop "Grid is not two dimensional, you should probably use getAlpha()!"
      end if

      result => this%alphaY
   end subroutine

   ! Gets the dimensions of the grid.
   !
   ! Returns: Dimensions, either 1 or 2.
   subroutine getGridDim(this, result)
      type(grid_t), intent(in) :: this
      integer, intent(out) :: result

      result = this%dim
   end subroutine

   ! Gets length of 1D grid. Must be one dimensional grid.
   !
   ! Returns: Length of 1D grid.
   subroutine getGridLength(this, result)
      type(grid_t), intent(in) :: this
      integer, intent(out) :: result

      if (this%dim .ne. 1) then
         error stop "Grid is not one dimensional, you should probably use getRow() or getCol()!"
      end if

      result = this%col
   end subroutine

   ! Gets the number of rows of the grid.
   !
   ! Returns: Number of rows.
   subroutine getGridRow(this, result)
      type(grid_t), intent(in) :: this
      integer, intent(out) :: result

      result = this%row
   end subroutine

   ! Gets the number of columns of the grid.
   !
   ! Returns: Number of columns.
   subroutine getGridCol(this, result)
      type(grid_t), intent(in) :: this
      integer, intent(out) :: result

      result = this%col
   end subroutine

   ! Sets the domain length of a 1D-Grid. Grid must be one dimensional.
   !
   ! Arguments:
   !   domainLength: A double value of the domain length. Must be positive.
   subroutine setGridDomain1D(this, domainLength)
      type(grid_t), intent(inout) :: this
      real(real64), intent(in) :: domainLength

      if (this%dim .ne. 1) then
         error stop "Grid is not one dimensional, you should probaly use the 2D domain setter!"
      end if

      if (domainLength .le. 0) then
         error stop "Given domain length is not positive!"
      end if

      this%domainCol = domainLength
      this%deltaCol = this%domainCol / this%col
   end subroutine

   ! Sets the domain size of a 2D-Grid. Grid must be two dimensional.
   !
   ! Arguments:
   !   domainRow: A double value of the domain size in y-direction. Must be positive.
   !   domainCol: A double value of the domain size in x-direction. Must be positive.
   subroutine setGridDomain2D(this, domainRow, domainCol)
      type(grid_t), intent(inout) :: this
      real(real64), intent(in) :: domainRow
      real(real64), intent(in) :: domainCol

      if (this%dim .ne. 2) then
         error stop "Grid is not two dimensional, you should probably use the 1D domain setter!"
      end if

      if ((domainRow .le. 0) .or. (domainCol .le. 0)) then
         error stop "Given domain size is not positive!"
      end if

      this%domainCol = domainCol
      this%domainRow = domainRow

      this%deltaCol = this%domainCol / this%col
      this%deltaRow = this%domainRow / this%row
   end subroutine

   ! Gets the delta value for 1D-Grid. Grid must be one dimensional.
   !
   ! Returns: Delta value.
   subroutine getGridDelta(this, result)
      type(grid_t), intent(in) :: this
      real(real64), intent(out) :: result

      if (this%dim .ne. 1) then
         error stop "Grid is not one dimensional, you should probably use the 2D delta getters"
      end if

      result = this%deltaCol
   end subroutine

   ! Gets the delta value in x-direction.
   !
   ! Returns: Delta value in x-direction.
   subroutine getGridDeltaCol(this, result)
      type(grid_t), intent(in) :: this
      real(real64), intent(out) :: result

      result = this%deltaCol
   end subroutine

   ! Gets the delta value in y-direction. Must be two dimensional grid.
   !
   ! Returns: Delta value in y-direction.
   subroutine getGridDeltaRow(this, result)
      type(grid_t), intent(in) :: this
      real(real64), intent(out) :: result

      if (this%dim .ne. 2) then
         error stop "Grid is not two dimensional, meaning there is no delta in y-direction!"
      end if

      result = this%deltaRow
   end subroutine
end module
