module simulation
   use, intrinsic :: iso_c_binding
   use, intrinsic :: iso_fortran_env

   use btcs

   use boundary
   use grid

   implicit none

   ! Enum holding different options for .csv output.
   enum, bind(C)
      enumerator :: CSV_OUTPUT_OFF      ! do not produce csv output
      enumerator :: CSV_OUTPUT_ON       ! produce csv output with last concentration matrix
      enumerator :: CSV_OUTPUT_VERBOSE  ! produce csv output with all concentration matrices
      enumerator :: CSV_OUTPUT_XTREME   ! csv output like VERBOSE but additional boundary conditions at beginning
   end enum

   ! Enum holding different options for console output.
   enum, bind(C)
      enumerator :: CONSOLE_OUTPUT_OFF     ! do not print any output to console
      enumerator :: CONSOLE_OUTPUT_ON      ! print before and after concentrations to console
      enumerator :: CONSOLE_OUTPUT_VERBOSE ! print all concentration matrices to console
   end enum

   integer, parameter :: CSV_OUTPUT = kind(CSV_OUTPUT_XTREME)
   integer, parameter :: CONSOLE_OUTPUT = kind(CONSOLE_OUTPUT_VERBOSE)

   ! The class forms the interface for performing the diffusion simulations and contains all the
   ! methods for controlling the desired parameters, such as time step, number of simulations, etc.
   type :: simulation_t
      real(real64) :: timestep
      integer :: iterations
      integer :: innerIterations
      integer(CSV_OUTPUT) :: csv_output_mode
      integer(CONSOLE_OUTPUT) :: console_output_mode

      type(grid_t), pointer :: grid
      type(boundary_t), pointer :: bc
   end type
contains
   ! Set up a simulation environment. The timestep and number of iterations must be set. For the
   ! BTCS approach, the Thomas algorithm is used as the default linear equation solver as this is
   ! faster for tridiagonal matrices. CSV output, console output and time measure are off by default.
   !
   ! Arguments:
   !   grid: Valid grid object
   !   bc: Valid boundary condition object
   subroutine initSimulation(this, grid, bc)
      type(simulation_t), intent(out) :: this
      type(grid_t), intent(in), pointer :: grid
      type(boundary_t), intent(in), pointer :: bc

      this%timestep = -1
      this%iterations = -1
      this%innerIterations = 1
      this%csv_output_mode = CSV_OUTPUT_OFF
      this%console_output_mode = CONSOLE_OUTPUT_OFF
      this%grid => grid
      this%bc => bc
   end subroutine

   ! Set the option to output the results to a CSV file. Off by default.
   !
   ! Arguments:
   !   csv_output: Valid output option.
   subroutine setSimulationOutputCSV(this, csv_output_mode)
      type(simulation_t), intent(inout) :: this
      integer(CSV_OUTPUT), intent(in) :: csv_output_mode

      this%csv_output_mode = csv_output_mode
   end subroutine

   ! Set the option to output the results to a CSV file. Off by default.
   !
   ! Arguments:
   !   csv_output: Valid output option.
   subroutine setSimulationOutputConsole(this, console_output_mode)
      type(simulation_t), intent(inout) :: this
      integer(CONSOLE_OUTPUT), intent(in) :: console_output_mode

      this%console_output_mode = console_output_mode
   end subroutine

   ! Setting the time step for each iteration step. Time step must be greater than zero.
   ! Setting the timestep is required.
   !
   ! Arguments:
   !   timestep: Valid timestep greater than zero.
   subroutine setSimulationTimestep(this, timestep)
      type(simulation_t), intent(inout) :: this
      real(real64), intent(in) :: timestep

      if (timestep <= 0) then
         error stop "Timestep has to be greater than zero."
      end if

      this%timestep = timestep
   end subroutine

   ! Currently set time step is returned.
   subroutine getSimulationTimestep(this, result)
      type(simulation_t), intent(in) :: this
      real(real64), intent(out) :: result

      result = this%timestep
   end subroutine

   ! Set the desired iterations to be calculated. A value greater than zero must be specified here.
   ! Setting iterations is required.
   !
   ! Arguments:
   !   iterations: Number of iterations to be simulated.
   subroutine setSimulationIterations(this, iterations)
      type(simulation_t), intent(inout) :: this
      integer, intent(in) :: iterations

      if (iterations <= 0) then
         error stop "Number of iterations must be greater than zero."
      end if

      this%iterations = iterations
   end subroutine

   ! Return the currently set iterations to be calculated.
   subroutine getSimulationIterations(this, result)
      type(simulation_t), intent(in) :: this
      integer, intent(out) :: result

      result = this%iterations
   end subroutine

   ! Outputs the current concentrations of the grid on the console.
   subroutine printConcentrationsConsole(this)
      type(simulation_t), intent(in) :: this
      real(real64), dimension(:, :), pointer :: concentrations

      call getGridConcentrations(this%grid, concentrations)

      print *, concentrations
      print *, ""
   end subroutine

   ! Creates a CSV file with a name containing the current simulation parameters.
   ! If the data name already exists, an additional counter is appended to the name.
   ! The name of the file is built up as follows:
   ! <approach> + <number rows> + <number columns> + <number of iterations>+<counter>.csv
   !
   ! Returns: Filename with configured simulation parameters.
   subroutine createCSVFile(this, filename)
      type(simulation_t), intent(in) :: this
      character(256), intent(out) :: filename

      integer :: appendIdent
      character(8) :: appendIdentString

      integer :: gridRows
      integer :: gridCols

      character(8) :: row
      character(8) :: col
      character(8) :: numIterations

      logical :: exists
      integer :: file

      real(real64), dimension(:), allocatable :: side
      integer(BC_SIDE), dimension(4), parameter :: sides = [BC_SIDE_LEFT, BC_SIDE_RIGHT, BC_SIDE_TOP, BC_SIDE_BOTTOM]

      integer :: i
      integer :: j

      appendIdent = 0
      appendIdentString = ""

      call getGridRow(this%grid, gridRows)
      call getGridCol(this%grid, gridCols)

      write(row, "(I0.0)") gridRows
      write(col, "(I0.0)") gridCols
      write(numIterations, "(I0.0)") this%iterations

      filename = row // "_" // col // "_" // numIterations // ".csv"
      inquire(file=filename, exist=exists)

      do while (exists)
         appendIdent = appendIdent + 1
         write(appendIdentString, "(I0.0)") appendIdent

         filename = row // "_" // col // "_" // numIterations // "-" // appendIdentString // ".csv"
         inquire(file=filename, exist=exists)
      end do

      open(newunit=file, file=filename, status="new", action="write")

      ! adds lines at the beginning of verbose output csv that represent the boundary conditions
      ! and their values -1 in case of closed boundary
      if (this%csv_output_mode == CSV_OUTPUT_XTREME) then
         do i = 1, size(sides)
            call getBoundarySideValues(this%bc, sides(i), side)

            do j = 1, size(side, 1)
               write(file, "(ES0.6)", advance="no") side(j)

               if (j .lt. size(side, 1)) then
                  write(file, "(A)", advance="no") " "
               end if
            end do

            write(file, "(A)") ""
         end do

         write(file, "(A)") ""
         write(file, "(A)") ""
      end if

      close(file)
   end subroutine

   ! Writes the currently calculated concentration values of the grid into the CSV file with the passed filename.
   !
   ! Arguments:
   !   filename: Name of the file to which the concentration values are to be written.
   subroutine printConcentrationsCSV(this, filename)
      type(simulation_t), intent(in) :: this
      character(256), intent(in) :: filename

      integer :: file
      real(real64), dimension(:, :), pointer :: concentrations

      integer :: j
      integer :: i

      open(newunit=file, file=filename, status="old", position="append", action="write")

      call getGridConcentrations(this%grid, concentrations)

      do j = 1, size(concentrations, 2)
         do i = 1, size(concentrations, 1)
            write(file, "(ES0.6)", advance="no") concentrations(i, j)

            if (i < size(concentrations, 1)) then
               write(file, "(A)", advance="no") " "
            end if
         end do

         write(file, "(A)") ""
      end do

      write(file, "(A)") ""
      write(file, "(A)") ""

      close(file)
   end subroutine


   ! Method starts the simulation process with the previously set parameters.
   subroutine runSimulation(this)
      type(simulation_t), intent(inout) :: this

      character(256), allocatable :: filename
      integer :: i

      if (this%timestep == -1) then
         error stop "Timestep is not set!"
      end if

      if (this%iterations == -1) then
         error stop "Number of iterations are not set!"
      end if

      filename = ""

      if (this%console_output_mode .ne. CONSOLE_OUTPUT_OFF) then
         call printConcentrationsConsole(this)
      end if

      if (this%csv_output_mode .ne. CSV_OUTPUT_OFF) then
         call createCSVFile(this, filename)
      end if

      do i = 1, this%iterations
         if (this%console_output_mode .eq. CONSOLE_OUTPUT_VERBOSE .and. i .lt. 1) then
            call printConcentrationsConsole(this)
         end if

         if ((this%csv_output_mode .eq. CSV_OUTPUT_VERBOSE) .or. (this%csv_output_mode .eq. CSV_OUTPUT_XTREME)) then
            call printConcentrationsCSV(this, filename)
         end if

         call BTCS_Thomas(this%grid, this%bc, this%timestep)
      end do

      if (this%console_output_mode .ne. CONSOLE_OUTPUT_OFF) then
         call printConcentrationsConsole(this)
      end if

      if (this%csv_output_mode .ne. CSV_OUTPUT_OFF) then
         call printConcentrationsCSV(this, filename)
      end if
   end subroutine
end module
