# Enum defining the two implemented boundary conditions.
@enum BC_TYPE begin
    BC_TYPE_CLOSED
    BC_TYPE_CONSTANT
end

# Enum defining all 4 possible sides to a 1D and 2D grid.
@enum BC_SIDE begin
    BC_SIDE_LEFT = 1
    BC_SIDE_RIGHT = 2
    BC_SIDE_TOP = 3
    BC_SIDE_BOTTOM = 4
end

# This class defines the boundary conditions of individual boundary elements.
# These can be flexibly used and combined later in other classes.
# The class serves as an auxiliary class for structuring the Boundary class.
#
# T is the data type of the boundary condition element
mutable struct BoundaryElement{T<:Number}
    type::BC_TYPE
    value::T

    # Construct a new Boundary Element object for the closed case. The boundary type is here
    # automatically set to the type BC_TYPE_CLOSED, where the value takes -1 and does not hold
    # any physical meaning.
    function BoundaryElement{T}() where {T<:Number}
        new(BC_TYPE_CLOSED, -1)
    end

    # Construct a new Boundary Element object for the constant case. The boundary type is
    # automatically set to the type BC_TYPE_CONSTANT.
    #
    # Arguments:
    #   value: Value of the constant concentration to be assumed at the corresponding
    #          boundary element.
    function BoundaryElement{T}(value::T) where {T<:Number}
        new(BC_TYPE_CONSTANT, value)
    end
end

# Allows changing the boundary type of a corresponding BoundaryElement object.
#
# Arguments:
#   type: Type of boundary condition. Either BC_TYPE_CONSTANT or BC_TYPE_CLOSED.
function setBoundaryElementType!(this::BoundaryElement{T}, type::BC_TYPE) where {T}
    this.type = type
end

# Sets the value of a boundary condition for the constant case.
#
# Arguments:
#   value: Concentration to be considered constant for the corresponding boundary element.
function setBoundaryElementValue!(this::BoundaryElement{T}, value::T) where {T}
    if this.type == BC_TYPE_CLOSED
        error(
            "No constant boundary concentrations can be set for closed boundaries. Please change type first.",
        )
    end

    this.value = value
end

# Return the type of the boundary condition, i.e. whether the boundary is considered closed or constant.
#
# Returns: Type of boundary condition, either BC_TYPE_CLOSED or BC_TYPE_CONSTANT.
function getBoundaryElementType(this::BoundaryElement{T})::BC_TYPE where {T}
    return this.type
end

# Return the concentration value for the constant boundary condition.
#
# Returns: Value of the concentration.
function getBoundaryElementValue(this::BoundaryElement{T})::T where {T}
    return this.value
end

mutable struct Boundary{T<:Number}
    dim::UInt8
    cols::UInt32
    rows::UInt32

    # Vector with Boundary Element information
    boundaries::Vector{Vector{BoundaryElement{T}}}

    # Creates a boundary object for a 1D grid
    #
    # Arguments:
    #   length: Length of the grid
    function Boundary{T}(length::UInt32) where {T}
        Boundary{T}(Grid{T}(Int(length)))
    end

    # Creates a boundary object for a 2D grid
    #
    # Arguments:
    #   rows: Number of rows of the grid
    #   cols: Number of columns of the grid
    function Boundary{T}(rows::UInt32, cols::UInt32) where {T}
        Boundary{T}(Grid{T}(Int(rows), Int(cols)))
    end

    # Creates a boundary object based on the passed grid object and initializes the boundaries as closed.
    #
    # Arguments:
    #   grid: Grid object on the basis of which the simulation takes place and from which the
    #         dimensions (in 2D case) are taken.
    function Boundary{T}(grid::Grid{T}) where {T}
        dim::UInt8 = getGridDim(grid)
        cols::UInt32 = getGridCol(grid)
        rows::UInt32 = getGridRow(grid)

        boundaries = Vector{Vector{BoundaryElement{T}}}(undef, 4)

        if dim == 1
            # in 1D only left and right boundary
            boundaries[Integer(BC_SIDE_LEFT)] = [BoundaryElement{T}()]
            boundaries[Integer(BC_SIDE_RIGHT)] = [BoundaryElement{T}()]
        elseif dim == 2
            boundaries[Integer(BC_SIDE_LEFT)] = [BoundaryElement{T}() for _ = 1:rows]
            boundaries[Integer(BC_SIDE_RIGHT)] = [BoundaryElement{T}() for _ = 1:rows]
            boundaries[Integer(BC_SIDE_TOP)] = [BoundaryElement{T}() for _ = 1:cols]
            boundaries[Integer(BC_SIDE_BOTTOM)] = [BoundaryElement{T}() for _ = 1:cols]
        end

        new(dim, cols, rows, boundaries)
    end
end

# Sets all elements of the specified boundary side to the boundary condition closed.
#
# Arguments:
#   side: Side to be set to closed, e.g. BC_SIDE_LEFT.
function setBoundarySideClosed!(this::Boundary{T}, side::BC_SIDE) where {T}
    if this.dim == 1
        if side == BC_SIDE_BOTTOM || side == BC_SIDE_TOP
            error(
                "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist.",
            )
        end
    end

    is_vertical::Bool = side == BC_SIDE_LEFT || side == BC_SIDE_RIGHT
    n::Int = is_vertical ? this.rows : this.cols

    this.boundaries[Integer(side)] = [BoundaryElement{T}() for _ = 1:n]
end

# Sets all elements of the specified boundary side to the boundary condition constant.
# Thereby the concentration values of the boundaries are set to the passed value.
#
# Arguments:
#   side: Side to be set to constant, e.g. BC_SIDE_LEFT.
#   value: Concentration to be set for all elements of the specified page.
function setBoundarySideConstant!(this::Boundary{T}, side::BC_SIDE, value::T) where {T}
    if this.dim == 1
        if side == BC_SIDE_BOTTOM || side == BC_SIDE_TOP
            error(
                "For the one-dimensional case, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist.",
            )
        end
    end

    is_vertical::Bool = side == BC_SIDE_LEFT || side == BC_SIDE_RIGHT
    n::Int = is_vertical ? this.rows : this.cols

    this.boundaries[Integer(side)] = [BoundaryElement{T}(value) for _ = 1:n]
end

# Specifically sets the boundary element of the specified side defined by the index to the
# boundary condition closed.
#
# Arguments:
#   side: Side in which an element is to be defined as closed.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
function setBoundaryElementClosed!(this::Boundary{T}, side::BC_SIDE, index::Int) where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[Integer(side)], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    setBoundaryElementType!(this.boundaries[Integer(side)][index], BC_TYPE_CLOSED)
end

# Specifically sets the boundary element of the specified side defined by the index to the boundary
# condition constant with the given concentration value.
#
# Arguments:
#   side: Side in which an element is to be defined as constant.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
#   value: Concentration value to which the boundary element should be set.
function setBoundaryElementConstant!(
    this::Boundary{T},
    side::BC_SIDE,
    index::Int,
    value::T,
) where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[Integer(side)], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    setBoundaryElementType!(this.boundaries[Integer(side)][index], BC_TYPE_CONSTANT)
    setBoundaryElementValue!(this.boundaries[Integer(side)][index], value)
end

# Returns the boundary condition of a specified side as a vector of BoundarsElement objects.
#
# Arguments:
#   side: Boundary side from which the boundary conditions are to be returned.
# Returns: Contains the boundary conditions as BoundaryElement<T> objects.
function getBoundarySide(
    this::Boundary{T},
    side::BC_SIDE,
)::AbstractVector{BoundaryElement{T}} where {T}
    if this.dim == 1
        if side == BC_SIDE_BOTTOM || side == BC_SIDE_TOP
            error(
                "For the one-dimensional trap, only the BC_SIDE_LEFT and BC_SIDE_RIGHT borders exist.",
            )
        end
    end

    return this.boundaries[Integer(side)]
end

# Returns the boundary condition of a specified element on a given side.
#
# Arguments:
#   side: Boundary side in which the boundary condition is located.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
# Return: Boundary condition as a BoundaryElement{T} object.
function getBoundaryElement(
    this::Boundary{T},
    side::BC_SIDE,
    index::Int,
)::BoundaryElement{T} where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[Integer(side)], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    return this.boundaries[Integer(side)][index]
end

# Returns the type of a boundary condition, i.e. either BC_TYPE_CLOSED or BC_TYPE_CONSTANT.
#
# Arguments:
#   side: Boundary side in which the boundary condition is located.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
# Return: Boundary Type of the corresponding boundary condition.
function getBoundaryElementType(this::Boundary{T}, side::BC_SIDE, index::Int)::BC_TYPE where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[Integer(side)], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    return getBoundaryElementType(this.boundaries[Integer(side)][index])
end

# Returns the concentration value of a corresponding BoundaryElement{T}
# object if it is a constant boundary condition.
#
# Arguments:
#   side: Boundary side in which the boundary condition is located.
#   index: Index of the boundary element on the corresponding boundary side.
#          Must index an element of the corresponding side.
# Return: Concentration of the corresponding BoundaryElement{T} object.
function getBoundaryElementValue(this::Boundary{T}, side::BC_SIDE, index::Int)::T where {T}
    # tests whether the index really points to an element of the boundary side.
    if (size(this.boundaries[Integer(side)], 1) < index) || index < 1
        error("Index is selected either too large or too small.")
    end

    return getBoundaryElementValue(this.boundaries[Integer(side)][index])
end

# Get thes Boundary Side Values as a vector. Value is -1 in case some specific boundary is closed.
#
# Arguments:
#   side: Boundary side for which the values are to be returned.
# Returns: Vector with values as T.
function getBoundarySideValues(this::Boundary{T}, side::BC_SIDE)::AbstractVector{T} where {T}
    length::Int = size(this.boundaries[Integer(side)], 1)
    values = Vector{T}(undef, length)

    for i = 1:length
        if getBoundaryElementType(this, side, i) == BC_TYPE_CLOSED
            values[i] = -1
            continue
        end

        values[i] = getBoundaryElementValue(this, side, i)
    end

    return values
end
