using FLoops
using LinearAlgebra

mutable struct Diagonals{T<:Real}
    left::Vector{T}
    center::Vector{T}
    right::Vector{T}

    function Diagonals{T}(size::Int) where {T<:Real}
        new(Vector{T}(undef, size), Vector{T}(undef, size), Vector{T}(undef, size))
    end
end

# calculates coefficient for boundary in constant case
function calcBoundaryCoeffConstant(
    alpha_center::T,
    alpha_side::T,
    sx::T,
)::Tuple{T,T} where {T<:Number}
    centerCoeff::T = 1 + sx * (calcAlphaIntercell(alpha_center, alpha_side) + 2 * alpha_center)
    sideCoeff::T = -sx * calcAlphaIntercell(alpha_center, alpha_side)

    return (centerCoeff, sideCoeff)
end

# calculates coefficient for boundary in closed case
function calcBoundaryCoeffClosed(
    alpha_center::T,
    alpha_side::T,
    sx::T,
)::Tuple{T,T} where {T<:Number}
    centerCoeff::T = 1 + sx * calcAlphaIntercell(alpha_center, alpha_side)
    sideCoeff::T = -sx * calcAlphaIntercell(alpha_center, alpha_side)

    return (centerCoeff, sideCoeff)
end

# creates coefficient matrix for next time step from alphas in x-direction
function createCoeffMatrix(
    alpha::AbstractMatrix{T},
    bcLeft::AbstractVector{BoundaryElement{T}},
    bcRight::AbstractVector{BoundaryElement{T}},
    numCols::Int,
    rowIndex::Int,
    sx::T,
)::Diagonals{T} where {T}
    type::BC_TYPE = BC_TYPE_CLOSED
    diag::Diagonals{T} = Diagonals{T}(numCols)

    # left column
    type = getBoundaryElementType(bcLeft[rowIndex])

    if type == BC_TYPE_CONSTANT
        centerCoeffTop, rightCoeffTop =
            calcBoundaryCoeffConstant(alpha[1, rowIndex], alpha[2, rowIndex], sx)

        diag.center[1] = centerCoeffTop
        diag.right[1] = rightCoeffTop
    elseif type == BC_TYPE_CLOSED
        centerCoeffTop, rightCoeffTop =
            calcBoundaryCoeffClosed(alpha[1, rowIndex], alpha[2, rowIndex], sx)

        diag.center[1] = centerCoeffTop
        diag.right[1] = rightCoeffTop
    else
        error("Undefined Boundary Condition Type somewhere on Left or Top!")
    end

    # inner columns
    n::Int = numCols - 1

    for i = 2:n
        diag.left[i] = -sx * calcAlphaIntercell(alpha[i-1, rowIndex], alpha[i, rowIndex])
        diag.center[i] =
            1 +
            sx * (
                calcAlphaIntercell(alpha[i, rowIndex], alpha[i+1, rowIndex]) +
                calcAlphaIntercell(alpha[i-1, rowIndex], alpha[i, rowIndex])
            )
        diag.right[i] = -sx * calcAlphaIntercell(alpha[i, rowIndex], alpha[i+1, rowIndex])
    end

    # right column
    type = getBoundaryElementType(bcRight[rowIndex])

    if type == BC_TYPE_CONSTANT
        centerCoeffBottom, leftCoeffBottom =
            calcBoundaryCoeffConstant(alpha[numCols, rowIndex], alpha[numCols-1, rowIndex], sx)

        diag.left[numCols] = leftCoeffBottom
        diag.center[numCols] = centerCoeffBottom
    elseif type == BC_TYPE_CLOSED
        centerCoeffBottom, leftCoeffBottom =
            calcBoundaryCoeffClosed(alpha[numCols, rowIndex], alpha[numCols-1, rowIndex], sx)

        diag.left[numCols] = leftCoeffBottom
        diag.center[numCols] = centerCoeffBottom
    else
        error("Undefined Boundary Condition Type somewhere on Right or Bottom!")
    end

    return diag
end

# calculates explicit concentration at boundary in closed case
function calcExplicitConcentrationsBoundaryClosed(
    conc_center::T,
    alpha_center::T,
    alpha_neigbor::T,
    sy::T,
)::T where {T}
    return sy * calcAlphaIntercell(alpha_center, alpha_neigbor) * conc_center +
           (1 - sy * (calcAlphaIntercell(alpha_center, alpha_neigbor))) * conc_center
end

# calculates explicity concentration at boundary in constant case
function calcExplicitConcentrationsBoundaryConstant(
    conc_center::T,
    conc_bc::T,
    alpha_center::T,
    alpha_neighbor::T,
    sy::T,
)::T where {T}
    inter_cell::T = calcAlphaIntercell(alpha_center, alpha_neighbor)
    return sy * inter_cell * conc_center +
           (1 - sy * (inter_cell + alpha_center)) * conc_center +
           sy * alpha_center * conc_bc
end

function createSolutionVector(
    concentrations::AbstractMatrix{T},
    alphaX::AbstractMatrix{T},
    alphaY::AbstractMatrix{T},
    bcLeft::AbstractVector{BoundaryElement{T}},
    bcRight::AbstractVector{BoundaryElement{T}},
    bcTop::AbstractVector{BoundaryElement{T}},
    bcBottom::AbstractVector{BoundaryElement{T}},
    length::Int,
    rowIndex::Int,
    sx::T,
    sy::T,
)::AbstractVector{T} where {T}
    sv = Vector{T}(undef, length)
    numRows::Int = size(concentrations, 2)

    type::BC_TYPE = BC_TYPE_CLOSED

    # inner rows
    if rowIndex > 1 && rowIndex < numRows
        for i = 1:length
            sv[i] =
                sy *
                calcAlphaIntercell(alphaY[i, rowIndex], alphaY[i, rowIndex+1]) *
                concentrations[i, rowIndex+1] +
                (
                    1 -
                    sy * (
                        calcAlphaIntercell(alphaY[i, rowIndex], alphaY[i, rowIndex+1]) +
                        calcAlphaIntercell(alphaY[i, rowIndex-1], alphaY[i, rowIndex])
                    )
                ) * concentrations[i, rowIndex] +
                sy *
                calcAlphaIntercell(alphaY[i, rowIndex-1], alphaY[i, rowIndex]) *
                concentrations[i, rowIndex-1]
        end
    elseif rowIndex == 1 # first row
        for i = 1:length
            type = getBoundaryElementType(bcTop[i])

            if type == BC_TYPE_CONSTANT
                sv[i] = calcExplicitConcentrationsBoundaryConstant(
                    concentrations[i, rowIndex],
                    getBoundaryElementValue(bcTop[i]),
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex+1],
                    sy,
                )
            elseif type == BC_TYPE_CLOSED
                sv[i] = calcExplicitConcentrationsBoundaryClosed(
                    concentrations[i, rowIndex],
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex+1],
                    sy,
                )
            else
                error("Undefined Boundary Condition Type somewhere on Left or Top!")
            end
        end
    elseif rowIndex == numRows # last row
        for i = 1:length
            type = getBoundaryElementType(bcBottom[i])

            if type == BC_TYPE_CONSTANT
                sv[i] = calcExplicitConcentrationsBoundaryConstant(
                    concentrations[i, rowIndex],
                    getBoundaryElementValue(bcBottom[i]),
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex-1],
                    sy,
                )
            elseif type == BC_TYPE_CLOSED
                sv[i] = calcExplicitConcentrationsBoundaryClosed(
                    concentrations[i, rowIndex],
                    alphaY[i, rowIndex],
                    alphaY[i, rowIndex-1],
                    sy,
                )
            else
                error("Undefined Boundary Condition Type somewhere on Right or Bottom!")
            end
        end
    end

    # first column -> additional fixed concentration change from perpendicular dimension in constant bc case
    if getBoundaryElementType(bcLeft[rowIndex]) == BC_TYPE_CONSTANT
        sv[1] += 2 * sx * alphaX[1, rowIndex] * getBoundaryElementValue(bcLeft[rowIndex])
    end

    # last column -> additional fixed concentration change from perpendicular dimension in constant bc case
    if getBoundaryElementType(bcRight[rowIndex]) == BC_TYPE_CONSTANT
        sv[length] += 2 * sx * alphaX[length, rowIndex] * getBoundaryElementValue(bcRight[rowIndex])
    end

    return sv
end

# solver for linear equation system; A corresponds to coefficient matrix, b to the solution vector
# implementation of Thomas Algorithm
function ThomasAlgorithm(A::Diagonals{T}, x_vec::AbstractVector{T})::AbstractVector{T} where {T}
    n::Int = size(x_vec, 1)

    # start solving - c_diag and x_vec are overwritten
    n -= 1

    A.right[1] = A.right[1] / A.center[1]
    x_vec[1] = x_vec[1] / A.center[1]

    for i = 2:n
        A.right[i] /= A.center[i] - A.left[i] * A.right[i-1]
        x_vec[i] = (x_vec[i] - A.left[i] * x_vec[i-1]) / (A.center[i] - A.left[i] * A.right[i-1])
    end

    x_vec[n+1] = (x_vec[n+1] - A.left[n+1] * x_vec[n]) / (A.center[n+1] - A.left[n+1] * A.right[n])

    for i = 0:n-1
        x_vec[n-i] -= A.right[n-i] * x_vec[n-i+1]
    end

    return x_vec
end

# BTCS solution for 2D grid
function BTCS_2D!(grid::Grid{T}, bc::Boundary{T}, timestep::T) where {T}
    rowMax::Int = getGridRow(grid)
    colMax::Int = getGridCol(grid)

    sx::T = timestep / (2 * getGridDeltaCol(grid) * getGridDeltaCol(grid))
    sy::T = timestep / (2 * getGridDeltaRow(grid) * getGridDeltaRow(grid))

    concentrations_t1 = Matrix{T}(undef, (colMax, rowMax))

    alphaX::AbstractMatrix{T} = getGridAlphaX(grid)
    alphaY::AbstractMatrix{T} = getGridAlphaY(grid)

    bcLeft = getBoundarySide(bc, BC_SIDE_LEFT)
    bcRight = getBoundarySide(bc, BC_SIDE_RIGHT)
    bcTop = getBoundarySide(bc, BC_SIDE_TOP)
    bcBottom = getBoundarySide(bc, BC_SIDE_BOTTOM)

    concentrations::AbstractMatrix{T} = getGridConcentrations(grid)

    let alphaX = alphaX, alphaY = alphaY, concentrations_t1 = concentrations_t1

        @floop for i = 1:rowMax
            A = createCoeffMatrix(alphaX, bcLeft, bcRight, colMax, i, sx)
            b = createSolutionVector(
                concentrations,
                alphaX,
                alphaY,
                bcLeft,
                bcRight,
                bcTop,
                bcBottom,
                colMax,
                i,
                sx,
                sy,
            )

            concentrations_t1[:, i] = ThomasAlgorithm(A, b)
        end
    end

    concentrations_t1 = transpose(concentrations_t1)
    alphaX = transpose(alphaX)
    alphaY = transpose(alphaY)

    let alphaX = alphaX, alphaY = alphaY, concentrations_t1 = concentrations_t1

        @floop for i = 1:colMax
            # swap alphas, boundary conditions and sx/sy for column-wise calculation
            A = createCoeffMatrix(alphaY, bcTop, bcBottom, rowMax, i, sy)
            b = createSolutionVector(
                concentrations_t1,
                alphaY,
                alphaX,
                bcTop,
                bcBottom,
                bcLeft,
                bcRight,
                rowMax,
                i,
                sy,
                sx,
            )

            concentrations[i, :] = ThomasAlgorithm(A, b)
        end
    end
end

# differentiate between 1D and 2D grid
function BTCS_Thomas!(grid::Grid{T}, bc::Boundary{T}, timestep::T) where {T}
    if getGridDim(grid) == 2
        BTCS_2D!(grid, bc, timestep)
    else
        error("Error: Only 2-dimensional grids are defined!")
    end
end
