const MAT_INIT_VALUE::Number = 0

# Holds a matrix with concenctration and respective matrix/matrices of alpha coefficients.
#
# T is the type to be used for matrices, e.g. double or float
mutable struct Grid{T<:Real}
    col::Int
    row::Int
    dim::Int

    domainCol::T
    domainRow::T

    deltaCol::T
    deltaRow::T

    concentrations::AbstractMatrix{T}
    alphaX::AbstractMatrix{T}
    alphaY::AbstractMatrix{T}

    # Constructs a new 1D-Grid object of a given length, which holds a matrix with concentrations
    # and a respective matrix of alpha coefficients. The domain length is per default the same as
    # the length. The concentrations are all 20 by default and the alpha coefficients are 1.
    #
    # Arguments:
    #   length: Length of the 1D-Grid. Must be greater than 3.
    function Grid{T}(length::Int) where {T<:Real}
        if length <= 3
            error("Given grid length too small. Must be greater than 3.")
        end

        col::Int = length
        row::Int = 1

        dim::Int = 1

        domainCol::T = T(length)
        domainRow::T = 0

        deltaCol::T = domainCol / T(col) # -> 1
        deltaRow::T = 0

        concentrations::Matrix{T} = zeros(T, (col, 1))
        alphaX::Matrix{T} = zeros(T, (col, 1))
        alphaY::Matrix{T} = Matrix{T}(undef, (0, 0))

        fill!(concentrations, MAT_INIT_VALUE)
        fill!(alphaX, MAT_INIT_VALUE)

        new(col, row, dim, domainCol, domainRow, deltaCol, deltaRow, concentrations, alphaX, alphaY)
    end

    # Constructs a new 2D-Grid object of given dimensions, which holds a matrix with concentrations
    # and the respective matrices of alpha coefficient for each direction. The domain in
    # x- and y-direction is per default equal to the col length and row length, respectively. The
    # concentrations are all 20 by default across the entire grid and the alpha coefficients 1 in
    # both directions.
    #
    # Arguments:
    #   row: Length of the 2D-Grid in y-direction. Must be greater than 3.
    #   col: Length of the 2D-Grid in x-direction. Must be greater than 3.
    function Grid{T}(row::Int, col::Int) where {T<:Real}
        if row <= 1 || col <= 1
            error("At least one dimension is 1. Use 1D grid for better results.")
        end

        dim::Int = 2

        domainCol::T = T(col)
        domainRow::T = T(row)

        deltaRow::T = domainRow / T(row)
        deltaCol::T = domainCol / T(col)

        concentrations::Matrix{T} = Matrix{T}(undef, (col, row))
        alphaX::Matrix{T} = Matrix{T}(undef, (col, row))
        alphaY::Matrix{T} = Matrix{T}(undef, (col, row))

        fill!(concentrations, MAT_INIT_VALUE)
        fill!(alphaX, MAT_INIT_VALUE)
        fill!(alphaY, MAT_INIT_VALUE)

        new(col, row, dim, domainCol, domainRow, deltaCol, deltaRow, concentrations, alphaX, alphaY)
    end
end

const Grid32 = Grid{Float32}
const Grid64 = Grid{Float64}

# Sets the concentrations matrix for a 1D or 2D-Grid.
#
# Arguments:
#   concentrations: An Eigen3 MatrixX<T> holding the concentrations.
#                   Matrix must have correct dimensions as defined in row and col.
#                   (Or length, in 1D case).
function setGridConcentrations!(this::Grid{T}, concentrations::AbstractMatrix{T}) where {T}
    if size(concentrations) != size(this.concentrations)
        error("Given matrix of concentrations mismatch with Grid dimensions!")
    end

    this.concentrations = concentrations
end

# Gets the concentrations matrix for a Grid.
#
# Returns: A matrix holding the concentrations and having the same dimensions as the grid.
function getGridConcentrations(this::Grid{T})::AbstractMatrix{T} where {T}
    return this.concentrations
end

# Set the alpha coefficients of a 1D-Grid. Grid must be one  dimensional.
#
# Arguments
#   alpha: A matrix with 1 row holding the alpha coefficients. Matrix columns must have
#          same size as length of grid.
function setGridAlpha!(this::Grid{T}, alpha::AbstractMatrix{T}) where {T}
    if this.dim != 1
        error("Grid is not one dimensional, you should probably use 2D setter function!")
    end

    if size(alpha) != size(this.alphaX)
        error("Given matrix of alpha coefficients mismatch with Grid dimensions!")
    end

    this.alphaX = alpha
end

# Set the alpha coefficients of a 2D-Grid. Grid must be two dimensional.
#
# Arguments
#   alphaX: A matrix holding the alpha coefficients in x-direction. Matrix must be of
#           same size as the grid.
#   alphaY: A matrix holding the alpha coefficients in y-direction. Matrix must be of
#           same size as the grid.
function setGridAlpha!(
    this::Grid{T},
    alphaX::AbstractMatrix{T},
    alphaY::AbstractMatrix{T},
) where {T}
    if this.dim != 2
        error("Grid is not two dimensional, you should probably use 1D setter function!")
    end

    if size(alphaX) != size(this.alphaX)
        error("Given matrix of alpha coefficients in x-direction mismatch with Grid dimensions!")
    end

    if size(alphaY) != size(this.alphaY)
        error("Given matrix of alpha coefficients in y-direction mismatch with Grid dimensions!")
    end

    this.alphaX = alphaX
    this.alphaY = alphaY
end

# Gets the matrix of alpha coefficients of a 1D-Grid. Grid must be one dimensional.
#
# Returns: A matrix with 1 row holding the alpha coefficients.
function getGridAlpha(this::Grid{T})::AbstractMatrix{T} where {T}
    if this.dim != 1
        error(
            "Grid is not one dimensional, you should probably use either getAlphaX() or getAlphaY()!",
        )
    end

    return this.alphaX
end

# Gets the matrix of alpha coefficients in x-direction of a 2D-Grid. Grid must be two dimensional.
#
# Returns: A matrix holding the alpha coefficients in x-direction.
function getGridAlphaX(this::Grid{T})::AbstractMatrix{T} where {T}
    if this.dim != 2
        error("Grid is not two dimensional, you should probably use getAlpha()!")
    end

    return this.alphaX
end

# Gets the matrix of alpha coefficients in y-direction of a 2D-Grid. Grid must be two dimensional.
#
# Returns: A matrix holding the alpha coefficients in y-direction.
function getGridAlphaY(this::Grid{T})::AbstractMatrix{T} where {T}
    if this.dim != 2
        error("Grid is not two dimensional, you should probably use getAlpha()!")
    end

    return this.alphaY
end

# Gets the dimensions of the grid.
#
# Returns: Dimensions, either 1 or 2.
function getGridDim(this::Grid{T})::Int where {T}
    return this.dim
end

# Gets length of 1D grid. Must be one dimensional grid.
#
# Returns: Length of 1D grid.
function getGridLength(this::Grid{T})::Int where {T}
    if this.dim != 1
        error("Grid is not one dimensional, you should probably use getRow() or getCol()!")
    end

    return this.col
end

# Gets the number of rows of the grid.
#
# Returns: Number of rows.
function getGridRow(this::Grid{T})::Int where {T}
    return this.row
end

# Gets the number of columns of the grid.
#
# Returns: Number of columns.
function getGridCol(this::Grid{T})::Int where {T}
    return this.col
end

# Sets the domain length of a 1D-Grid. Grid must be one dimensional.
#
# Arguments:
#   domainLength: A double value of the domain length. Must be positive.
function setGridDomain!(this::Grid{T}, domainLength::T) where {T}
    if this.dim != 1
        error("Grid is not one dimensional, you should probaly use the 2D domain setter!")
    end

    if domainLength <= 0
        error("Given domain length is not positive!")
    end

    this.domainCol = domainLength
    this.deltaCol = this.domainCol / this.col
end

# Sets the domain size of a 2D-Grid. Grid must be two dimensional.
#
# Arguments:
#   domainRow: A double value of the domain size in y-direction. Must be positive.
#   domainCol: A double value of the domain size in x-direction. Must be positive.
function setGridDomain!(this::Grid{T}, domainRow::T, domainCol::T) where {T}
    if this.dim != 2
        error("Grid is not two dimensional, you should probably use the 1D domain setter!")
    end

    if domainRow <= 0 || domainCol <= 0
        error("Given domain size is not positive!")
    end

    this.domainRow = domainRow
    this.domainCol = domainCol
    this.deltaRow = this.domainRow / this.row
    this.deltaCol = this.domainCol / this.col
end

# Gets the delta value for 1D-Grid. Grid must be one dimensional.
#
# Returns: Delta value.
function getGridDelta(this::Grid{T})::T where {T}
    if this.dim != 1
        error("Grid is not one dimensional, you should probably use the 2D delta getters")
    end

    return this.deltaCol
end

# Gets the delta value in x-direction.
#
# Returns: Delta value in x-direction.
function getGridDeltaCol(this::Grid{T})::T where {T}
    return this.deltaCol
end

# Gets the delta value in y-direction. Must be two dimensional grid.
#
# Returns: Delta value in y-direction.
function getGridDeltaRow(this::Grid{T})::T where {T}
    if this.dim != 2
        error("Grid is not two dimensional, meaning there is no delta in y-direction!")
    end

    return this.deltaRow
end
